# hook to load information from tally configuration

import os
import sys
import warnings
import importlib.util
from pathlib import Path

URL = "https://github.com/seanbreckenridge/ttally"


def get_conf_file() -> Path:
    cfg_file: str = os.environ.get("TTALLY_CFG", "~/.config/ttally.py")
    cfg_path = Path(cfg_file).expanduser().absolute()
    if not cfg_path.parent.exists():
        warnings.warn("Directory for configuration doesn't exist, creating...")
        cfg_path.parent.mkdir()
    if not cfg_path.exists():
        raise FileNotFoundError(
            f"Expected configuration to exist at {cfg_path}, see {URL} for an example"
        )
    return cfg_path


def setup_config() -> None:
    conf: Path = get_conf_file()
    sconf: str = str(conf)

    mname = "ttally.config"
    # redirect import
    if mname in sys.modules:
        del sys.modules[mname]
    try:
        spec = importlib.util.spec_from_file_location(mname, sconf)
        assert spec is not None
        mod = importlib.util.module_from_spec(spec)
        err = "Error importing configuration, must be on python3.5+?"
        assert spec.loader is not None, err
        # not sure why mypy is using the legacy (<3.4) types here?
        # https://docs.python.org/3/library/importlib.html#importlib.abc.Loader
        spec.loader.exec_module(mod)  # type: ignore[attr-defined]
        if mname not in sys.modules:
            sys.modules[mname] = mod
        # make sure its importable
        import ttally.config  # noqa
    except ImportError as e:
        raise ImportError(
            f"""
Importing 'ttally.config' failed! (error: {e}).
This file is typically located at  ~/.config/ttally.py, see {URL} for
more information
"""
        )


setup_config()
del setup_config



"""
Global functions which glob all data from the TTALLY_DATA_DIR
Used with 'hpi query' https://github.com/karlicoss/HPI
to access this from the command line

E.g. to grab how many things I've eaten within the last day
$ hpi query ttally.funcs.food --recent 1d | jq length
25
"""

from typing import Any

from .autotui_ext import glob_namedtuple
from .models import MODELS


# some metaprogramming hackery to create functions at the global level
# this intercepts any AttributeErrors thrown at a global level -- if the name
# you're trying to access/import matches a model name, it creates a function
# which returns the data for that


def __getattr__(name: str) -> Any:
    if name in MODELS:
        return lambda: glob_namedtuple(MODELS[name])
    raise AttributeError(f"No such attribute {name}")



import os
from pathlib import Path
from datetime import datetime
from typing import NamedTuple, Iterator, Optional, Type, List, Dict, Any, TextIO
from itertools import chain

from autotui import namedtuple_sequence_loads, prompt_namedtuple
from autotui.shortcuts import load_prompt_and_writeback, load_from, dump_to

from .file import datafile, glob_datafiles


def namedtuple_func_name(nt: Type[NamedTuple]) -> str:
    assert hasattr(nt, "_fields"), "Did not receive a valid NamedTuple!"
    return str(nt.__name__.casefold())


# load, prompt and writeback one of the models
def prompt(nt: Type[NamedTuple]) -> None:
    f: Path = datafile(namedtuple_func_name(nt))
    load_prompt_and_writeback(nt, f)


# prompt, but set the datetime for the resulting nametuple to now
def prompt_now(nt: Type[NamedTuple]) -> None:
    # load items from file
    p: Path = datafile(namedtuple_func_name(nt))
    load_prompt_and_writeback(nt, p, type_use_values={datetime: datetime.now})


# takes one of the models.py and loads all data from it
def glob_namedtuple(
    nt: Type[NamedTuple], in_dir: Optional[Path] = None
) -> Iterator[NamedTuple]:
    yield from chain(
        *map(
            lambda p: load_from(nt, p, allow_empty=True),
            glob_datafiles(namedtuple_func_name(nt), in_dir=in_dir),
        )
    )


# used in __main__.py for the from_json command
def save_from(nt: Type[NamedTuple], use_input: TextIO, partial: bool = False) -> None:
    json_text: str = use_input.read()
    p = datafile(namedtuple_func_name(nt))
    items: List[NamedTuple] = load_from(nt, p, allow_empty=True)
    new_items: List[NamedTuple] = []
    if partial:
        # load the list as json blobs
        os.environ["AUTOTUI_DISABLE_WARNINGS"] = "1"  # ignore null warnings
        blobs: List[Dict[str, Any]] = []
        for b in namedtuple_sequence_loads(json_text, nt):
            blobs.append({k: v for k, v in b._asdict().items() if v is not None})
        del os.environ["AUTOTUI_DISABLE_WARNINGS"]
        for bd in blobs:
            new_nt = prompt_namedtuple(nt, attr_use_values=bd)
            new_items.append(new_nt)
    else:
        new_items.extend(namedtuple_sequence_loads(json_text, nt))
    items.extend(new_items)
    dump_to(items, p)



from typing import List, NamedTuple, Type, Callable, Any, Iterator
from datetime import datetime

import more_itertools

from .autotui_ext import glob_namedtuple


def namedtuple_extract_from_annotation(nt: Type[NamedTuple], _type: Any) -> str:
    """
    >>> from typing import NamedTuple; from datetime import datetime
    >>> class Test(NamedTuple): something: datetime
    >>> namedtuple_extract_from_annotation(Test, datetime)
    'something'
    """
    import inspect
    from autotui.typehelpers import resolve_annotation_single

    for attr_name, param in inspect.signature(nt).parameters.items():
        # Optional[(<class 'int'>, False)]
        attr_type, _ = resolve_annotation_single(param.annotation)

        if attr_type == _type:
            return attr_name
    raise TypeError(f"Could not find {_type} on {nt}")


def _extract_dt_from(nt: Type[NamedTuple]) -> Callable[[NamedTuple], datetime]:
    # returns a function, when which given an item of this
    # type, returns the datetime value
    dt_attr: str = namedtuple_extract_from_annotation(nt, datetime)
    return lambda o: getattr(o, dt_attr)  # type: ignore[no-any-return]


def sort_namedtuple_by_date(nt: Type[NamedTuple], reverse: bool = False) -> List[NamedTuple]:
    return sorted(glob_namedtuple(nt), key=_extract_dt_from(nt), reverse=reverse)


def query_recent(nt: Type[NamedTuple], count: int) -> List[NamedTuple]:
    """query the module for recent entries (based on datetime) from a namedtuple"""
    items: List[NamedTuple] = more_itertools.take(
        count,
        sort_namedtuple_by_date(nt, reverse=True)
    )
    return items


def query_print(nt: Type[NamedTuple], count: int, remove_attrs: List[str]) -> None:
    # assumes that there is a datetime attribute on this, else
    # we have nothing to sort by
    res = more_itertools.peekable(iter(query_recent(nt, count)))
    try:
        first_item = res.peek()  # namedtuple-like
    except StopIteration:
        raise RuntimeError(f"data queried from {nt} was empty")
    dt_attr: str = namedtuple_extract_from_annotation(first_item.__class__, datetime)
    # get non-datetime attr names, if they're not filtered
    other_attrs: List[str] = [
        k for k in first_item._asdict().keys() if k != dt_attr and k not in remove_attrs
    ]
    for o in res:
        print(datetime.fromtimestamp(getattr(o, dt_attr).timestamp()), end="\t")
        print(" \t".join([str(getattr(o, a)) for a in other_attrs]))



from typing import Iterator

from .models import MODELS


def generate_shell_aliases(python_loc: str = "python3") -> Iterator[str]:
    pre = f"'{python_loc} -m ttally "
    suf = "'"
    for mname, model in MODELS.items():
        yield f"alias {mname}={pre}prompt {mname}{suf}"
        yield f"alias {mname}-now={pre}prompt-now {mname}{suf}"
        yield f"alias {mname}-recent={pre}recent {mname}{suf}"



import sys
import json
from typing import NamedTuple, Type, Optional, List, Sequence

import click

from .models import MODELS


def _model_from_string(model_name: str) -> Type[NamedTuple]:
    try:
        return MODELS[model_name]
    except KeyError:
        click.echo(f"Could not find a model named {model_name}", err=True)
        sys.exit(1)


@click.group()
def main() -> None:
    """
    Tally things that I do often!

    Given a few namedtuples, this creates serializers/deserializers
    and an interactive interface using 'autotui', and aliases
    to:

    prompt using default autotui behavior, writing to the ttally datafile,
    same as above, but if the model has a datetime, set it to now,
    query the 10 most recent items for a model
    """
    pass


@main.command(short_help="generate shell aliases")
def generate() -> None:
    """
    Generate the shell aliases!
    """
    from .codegen import generate_shell_aliases

    for a in generate_shell_aliases():
        print(a)


def _model_names() -> List[str]:
    # sort this, so that the order doesn't change while tabbing through
    return sorted(m for m in MODELS)


def _model_complete(
    ctx: click.Context, args: Sequence[str], incomplete: str
) -> List[str]:
    return [m for m in _model_names() if m.startswith(incomplete)]


model_with_completion = click.argument("MODEL", shell_complete=_model_complete)


@main.command(short_help="add item by piping JSON")
@model_with_completion
@click.option(
    "-p",
    "--partial",
    default=False,
    is_flag=True,
    help="Allow partial input -- prompt any fields which arent provided",
)
@click.option(
    "-f",
    "--file",
    default=None,
    type=click.Path(exists=True),
    help="Read from file instead of STDIN",
)
def from_json(model: str, partial: bool, file: Optional[str]) -> None:
    """
    A way to allow external programs to save JSON data to the current file for the model

    Provide a list of JSON from STDIN, and the corresponding model to parse it to
    (in lowercase) as the first argument, and this parses (validates)
    and saves it to the file
    """
    from .autotui_ext import save_from

    if file is None:
        save_from(_model_from_string(model), use_input=sys.stdin, partial=partial)
    else:
        with open(file, "r") as f:
            save_from(_model_from_string(model), use_input=f, partial=partial)


@main.command(short_help="print the datafile location")
@model_with_completion
def datafile(model: str) -> None:
    """
    Print the location of the current datafile for some model
    """
    from .file import datafile as df

    m = model.lower()
    assert m in MODELS, f"Couldn't find model {m}"
    f = df(m)
    if not f.exists():
        click.secho(f"Warning: {f} doesn't exist", err=True, fg="red")
    click.echo(f)


@main.command(name="models", help="list models")
def _models_cmd() -> None:
    """
    List all ttally models
    """
    click.echo("\n".join(_model_names()))


@main.command(name="prompt", help="tally an item")
@model_with_completion
def _prompt(model: str) -> None:
    """
    Prompt for every field in the given model
    """
    from .autotui_ext import prompt

    prompt(_model_from_string(model))


@main.command(name="prompt-now", help="tally an item (now)")
@model_with_completion
def _prompt_now(model: str) -> None:
    """
    Prompt for every field in the model, except datetime, which should default to now
    """
    from .autotui_ext import prompt_now

    prompt_now(_model_from_string(model))


@main.command(name="recent", short_help="print recently tallied items")
@model_with_completion
@click.option(
    "-r",
    "--remove-attrs",
    type=str,
    default="",
    help="comma separated list of attributes to remove while printing",
)
@click.argument("COUNT", type=int, default=10)
def _recent(model: str, remove_attrs: str, count: int) -> None:
    """
    List recent items logged for this model
    """
    from .recent import query_print

    attrs = [a.strip() for a in remove_attrs.split(",") if a.strip()]
    query_print(_model_from_string(model), count, remove_attrs=attrs)


@main.command(short_help="export all data from a model")
@model_with_completion
@click.option(
    "-s",
    "--stream",
    default=False,
    is_flag=True,
    help="Stream objects as they're read, instead of a list",
)
def export(model: str, stream: bool) -> None:
    """
    List all the data from a model as JSON
    """
    from autotui.fileio import namedtuple_sequence_dumps
    from .autotui_ext import glob_namedtuple

    itr = json.loads(
        namedtuple_sequence_dumps(list(glob_namedtuple(_model_from_string(model))))
    )

    if stream:
        for blob in itr:
            sys.stdout.write(json.dumps(blob))
            sys.stdout.write("\n")
    else:
        sys.stdout.write(json.dumps(list(itr)))
        sys.stdout.write("\n")
    sys.stdout.flush()


@main.command(short_help="merge all data for a model into one file")
@model_with_completion
@click.option("--sort-key", default=None, help="Sort resulting merged data by JSON key")
def merge(model: str, sort_key: Optional[str]) -> None:
    """
    Merge all datafiles for one model into a single '-merged.json' file
    """
    from pathlib import Path
    from datetime import datetime

    from autotui.fileio import namedtuple_sequence_dumps
    from .autotui_ext import glob_namedtuple
    from .file import glob_datafiles, ttally_temp_dir, ttally_merged_path

    datafiles: List[Path] = list(glob_datafiles(model))
    if len(datafiles) == 0:
        click.echo(f"No datafiles for model {model}", err=True)
        return

    data = json.loads(
        namedtuple_sequence_dumps(list(glob_namedtuple(_model_from_string(model))))
    )

    # if provided, use sort key
    if sort_key is not None and len(data) > 0:
        assert sort_key in data[0], f"Could not find {sort_key} in {data[0]}"
        data = list(sorted(data, key=lambda obj: obj[sort_key]))  # type: ignore[no-any-return]

    epoch = int(datetime.now().timestamp())
    cachefile = ttally_temp_dir() / f"{model}-{epoch}-merged.json"

    click.echo(f"Writing backup to '{cachefile}'", err=True)
    with cachefile.open("w") as backup_f:
        json.dump(data, backup_f)

    # remove current datafiles
    for rmf in datafiles:
        click.echo(f"Removing '{rmf}'", err=True)
        rmf.unlink()

    merge_target = ttally_merged_path(model)
    with merge_target.open("w") as merged_f:
        json.dump(data, merged_f)

    click.echo(f"Wrote merged file to '{merge_target}'", err=True)


@main.command(short_help="edit the datafile")
@model_with_completion
def edit(model: str) -> None:
    """
    Edit the current datafile with your editor
    """
    from .file import datafile as df

    _model_from_string(model)
    f = df(model.lower())
    if not f.exists():
        click.secho(f"Warning: {f} doesn't exist", err=True, fg="red")
    click.edit(filename=str(f))


if __name__ == "__main__":
    main(prog_name="ttally")



import os
import sys
from functools import lru_cache
from pathlib import Path
from tempfile import gettempdir
from typing import Optional, Iterator


ENV = "TTALLY_DATA_DIR"
DEFAULT_DATA = "~/.local/share/ttally"


@lru_cache(1)
def versioned_timestamp() -> str:
    import socket
    from datetime import datetime

    OS = sys.platform.casefold()
    # for why this uses socket:
    # https://docs.python.org/3/library/os.html#os.uname
    HOSTNAME = "".join(socket.gethostname().split()).casefold()
    TIMESTAMP = datetime.strftime(datetime.now(), "%Y-%m")
    return f"{OS}-{HOSTNAME}-{TIMESTAMP}"


@lru_cache(1)
def ttally_abs() -> Path:
    ddir: str = os.environ.get(ENV, DEFAULT_DATA)
    p = Path(ddir).expanduser().absolute()
    if not p.exists():
        import warnings

        warnings.warn(f"{p} does not exist, creating...")
        p.mkdir()
    return p


def ttally_merged_path(model: str) -> Path:
    return ttally_abs() / f"{model}-merged.json"


# creates unique datafiles for each platform
def datafile(for_function: str, in_dir: Optional[Path] = None) -> Path:
    # add some OS/platform specific code to this, to prevent
    # conflicts across computers while using syncthing
    # this also decreases the amount of items that have
    # to be loaded into memory for load_prompt_and_writeback
    ext = os.environ.get("TTALLY_EXT", "yaml")
    u = f"{for_function}-{versioned_timestamp()}.{ext}"
    return Path(in_dir or ttally_abs()).absolute() / u


# globs all datafiles for some for_function
def glob_datafiles(for_function: str, in_dir: Optional[Path] = None) -> Iterator[Path]:
    d: Path = Path(in_dir or ttally_abs()).absolute()
    for f in os.listdir(d):
        if f.startswith(for_function):
            yield d / f


def ttally_cache_dir() -> Path:
    cache_dir = Path(os.environ.get("XDG_CACHE_DIR", str(Path.home() / ".cache")))
    ttally_cache_dir = cache_dir / "ttally"
    if not ttally_cache_dir.exists():
        ttally_cache_dir.mkdir(parents=True)
    return ttally_cache_dir


def ttally_temp_dir() -> Path:
    tdir = gettempdir()
    ttally_temp_dir = Path(tdir) / "ttally"
    if not ttally_temp_dir.exists():
        ttally_temp_dir.mkdir(parents=True)
    return ttally_temp_dir



import inspect

from typing import Any, Dict, NamedTuple, Type

# should already be configured, since __init__.py hook runs
# when the module is initially loaded

from ttally import config


def _is_model(o: Any) -> bool:
    return inspect.isclass(o) and issubclass(o, tuple) and hasattr(o, "_fields")


# dynamically create a list of each of these
MODELS: Dict[str, Type[NamedTuple]] = {
    name.casefold(): klass for name, klass in inspect.getmembers(config, _is_model)
}
